# -*- coding: utf-8 -*-
'''
dwave testing client GET/POST to agent via RESTFUL API, loaded json testing config file
testing-client.py
@copyright: (c) 2021 by CAICT
@modified: 2021-06-09
'''
import requests
import json
import time
import random

class Testing_Client_Dwave_Leap:
    def __init__(self):
        '''
        qc testing config parameters as json format
        @ tester: name of tester
        @ ip: ip address of agent server
        @ port: port number of agent server
        @ provider: provider of remote qc cloud platform via Internet
        @ operation_time: time of testing operation. e.g. 2021-06-09 09:40:30
        @ token: token to login qc cloud platform
        @ if_mq: if or not writing the recording results to MQ. e.g. datahub of aliyun
        @ query_type: type of qpu or simulator with qc provider. e.g. qpu_2000q is 2000qubit Quantum Annealing QPU provided by Dwave
        @ operation_item: type of testing operation. e.g. query runtime means checking the runtime of excuting qc task
        @ category_item: selected category of benchmarking framework designed by CAICT
        @ testing_epoch: epoch for testing or attempts
        '''
        self.config_params = {
            'tester': 'xx-caict',
            'ip': '8.140.123.60',
            'provider': 'dwave-leap',
            'operation_time': time.strftime('%Y/%m/%d %H:%M:%S'),
            'token': '*********your token*****************',
            'if_mq': 'True',
            'query_type': 'qpu_2000q',
            'operation_item': 'query_runtime',
            'category_item': 'operation_maintenance',
            'testing_epoch': '10',
        }
        print('successfully create instance of dwave leap testing client!!!!')
    
    def testing_params_config(self, config_params):
        keys = config_params.keys()
        for key in keys:
            self.config_params[key] = config_params[key]
        print('successfully config testing params of testing client!!!!!!')
        print(self.config_params)
        return self.config_params

    def save_config_file(self, config_params, filename = 'caict-testing-config.json'):
        self.config_params = config_params
        prex = time.strftime('%Y-%m-%d-%H-%M-%S-')
        with open(prex + filename, "w") as f:
            json.dump(self.config_params, f)
            print("successfully save config file!!!!!!!")

    def load_config_file(self, filename):
        with open(filename) as f:
            self.config_params = json.load(f)
            print('successfully load config file!!!!!')
            print(self.config_params)
        return self.config_params

    def login_agent_server(self):
        body = self.config_params
        url = 'http://' + body['ip'] + ':' + body['port'] + '/login'
        time1 = time.time()
        # ret = requests.post("http://8.140.123.60:1981/login", json.dumps(body))
        ret = requests.post(url, json.dumps(body))
        print('login time:')
        time2 = time.time()
        print("ELAPSED: ", time2 - time1)
        print("RESULT:", ret.json())
        return ret.json()

    def query_runtime(self):
        body = self.config_params
        url = 'http://' + body['ip'] + ':' + body['port'] + '/query_runtime/' + str(body['tester'])
        epoch = int(body['testing_epoch'])
        body['operation_time'] = time.strftime('%Y/%m/%d %H:%M:%S')
        if body['operation_item'] == 'query_runtime':
            for num in range(epoch):
                time1 = time.time()
                ret = requests.post(url, json.dumps(body))
                print('serving time of query_runtime:')
                time2 = time.time()
                print("ELAPSED: ", time2 - time1)
                print("RESULT:", ret.json())
        else:
            return {'query_status': False, 'debug_info': 'wrong testing param for query runtime:' + body['operation_item']}
        return ret.json()

    def query_solver(self):
        body = self.config_params
        url='http://' + body['ip'] + ':' + body['port'] + '/query_solver/' + str(body['tester'])
        epoch = int(body['testing_epoch'])
        body['operation_time'] = time.strftime('%Y/%m/%d %H:%M:%S')
        if body['operation_item'] == 'query_solver':
            for num in range(epoch):
                time1 = time.time()
                ret = requests.post(url, json.dumps(body))
                print('serving time of query_solver:')
                time2 = time.time()
                print("ELAPSED: ", time2 - time1)
        else:
            return {'query_status': False, 'debug_info': 'wrong testing param for query solver:' + body['operation_item']}
        return ret.json()
        # print("RESULT:", ret.json())

if __name__ == '__main__':
    config_body = {'tester': 'lvbo-caict',
            'ip': '8.140.123.60',
            'port': '1981',
            'provider': 'dwave-leap',
            'operation_time': time.strftime('%Y/%m/%d %H:%M:%S'),
            'token': '*********your token*****************',
            'if_mq': 'True',
            'query_type': 'qpu_2000q',
            'operation_item': 'query_runtime',
            'category_item': 'operation_maintenance',
            'testing_epoch': '3',
            }
    ## create instance of testing client for dwave_leap qc cloud platform
    dwave_client = Testing_Client_Dwave_Leap()
    ## config params of testing client
    dwave_client.testing_params_config(config_body)
    ## save config file as json
    dwave_client.save_config_file(config_body,'lvbo.json')
    ## load config params from json file
    dwave_client.query_runtime()



# -*- coding: utf-8 -*-
'''
dwave testing client GET/POST to agent via RESTFUL API, adopting argparse mode
'''
import argparse
import testing_client_json as client_module
import time
import logging
import sys
import json
FORMAT = '[%(asctime)s, %(levelname)-7s]: %(message)s'
logging.basicConfig(format = FORMAT)
logger = logging.getLogger('spider')
logger.setLevel(logging.INFO)

parser = argparse.ArgumentParser(description = 'caict testing qc benchmark platform project')
parser.add_argument('-v', '--verbose', help = 'debug info', action = 'store_true')
#group = parser.add_mutually_exclusive_group() # config manually or load config file is exclusive
parser.add_argument('-t', '--tester', help = 'name of tester', default = 'caict')
parser.add_argument('-i', '--ip', help = 'ip address of agent server', default = '8.140.123.60')
parser.add_argument('-p', '--port', help = 'port number of agent server', default = '1981')
parser.add_argument('-pr', '--provider', help = 'qc provider', default = 'dwave-leap')
parser.add_argument('-ti', '--operation_time', help = 'time of testing operation', default = time.strftime('%Y/%m/%d %H:%M:%S'))
parser.add_argument('-tk', '--token', help = 'login token of qc cloud platform', default = 'xxxxxxxxx your token xxxxxxxxxx')
parser.add_argument('-mq', '--if_mq', help = 'if or not recording testing results to MQ. e.g. datahub of aliyun', default = 'True')
parser.add_argument('-q', '--query_type', help = 'type of qpu or simulator', default = 'qpu_2000q')
parser.add_argument('-o', '--operation_item', help = 'type of testing operation', default = 'query_runtime')
parser.add_argument('-c', '--category_item', help = 'category benchmark designed by CAICT', default = 'operation_maintenance')
parser.add_argument('-e', '--testing_epoch', help = 'epoch of testing attemps repeadlly', default = '3')
parser.add_argument('-lc', '--load_config', help = 'load config from json file', default = '')
parser.add_argument('-sc', '--save_config', help = 'save config from json file', default = 'ignore')

args = parser.parse_args()

config_body = {}

if len(args.load_config) >0:
    print('==== load config file! ====')
    filename = args.load_config
    print(filename)
    with open(filename) as f:
        config_body = json.load(f)
elif args.load_config == '':
    config_body['tester'] = args.tester
    config_body['ip'] = args.ip
    config_body['port'] = args.port
    config_body['provider'] = args.provider
    config_body['operation_time'] = args.operation_time
    config_body['token'] = args.token
    config_body['if_mq'] = args.if_mq
    config_body['query_type'] = args.query_type
    config_body['operation_item'] = args.operation_item
    config_body['category_item'] = args.category_item
    config_body['testing_epoch'] = args.testing_epoch
else:
    print('======  error argument!!! ===========')

if args.save_config is not 'ignore':
    prex = time.strftime('%Y-%m-%d-%H-%M-%S-')
    with open(prex + args.save_config, 'w') as f:
        json.dump(config_body, f)
        print('successfully save config file!!!!!!!!!!')

# create instance of testing client for dwave_leap qc cloud platform
dwave_client = client_module.Testing_Client_Dwave_Leap()
## config params of testing client
dwave_client.testing_params_config(config_body)
## save config file as json
#dwave_client.save_config_file(config_body,'lvbo.json')
## load config params from json file
dwave_client.query_runtime()
time.sleep(2)





# -*- coding: utf-8 -*-
'''
dwave agent linking to Leap and provide testing service as an agent
'''
import requests
import uvicorn
from fastapi import FastAPI, Request
from pydantic import BaseModel
from datetime import datetime
from typing import List, Dict, Set, Union, Text, Tuple, Optional
from dwave_leap import Dwave_leap, MQ_Dwave_Leap_Runtime, MQ_Dwave_Leap_Sovler_property

class Item(BaseModel):
    tester: str = None
    operation_time: str = '1970-01-01 00:00:00'
    query_type: str = 'qpu_advantage'
    category_item: str = 'operation_maintenance'
    operation_item: str = 'query_runtime'
    if_mq: str = 'False'
    token: str


mq_dwave_solver = MQ_Dwave_Leap_Sovler_property()
mq_dwave_runtime = MQ_Dwave_Leap_Runtime()
app = FastAPI()

@app.get("/")
async  def root():
    return 'Hello dwave-leap QC cloud platform!'

@app.post("/login")
async  def login(item: Item):
    item_dict = item.dict()
    TOKEN = item_dict['token']
    # print(TOKEN)
    global dwave_leap
    dwave_leap = Dwave_leap(TOKEN)
    print('Hello dwave-leap QC cloud platform!!!!!!!!!!!!!!!!')
    return {"code": 200, "msg": "success", "login_state": True}

@app.post("/query_solver/{tester}")
async def query_solver(tester: str, item: Item):
    item_dict = item.dict()
    config_body = {}
    config_body['tester'] = str(tester)
    config_body['operation_time'] = item_dict['operation_time']
    config_body['query_type'] = item_dict['query_type']
    config_body['category_item'] = item_dict['category_item']
    config_body['operation_item'] = item_dict['operation_item']
    output_body = {}
    output_body = dwave_leap.query_solver_property(config_body['query_type'])
    res = output_body
    
    if item_dict['if_mq'] == 'True':
        mq_dwave_solver.mq_write(config_body, output_body)
        res['mq_write_state'] = True
    else:
        res['mq_write_state'] = False
    res['code'] = 200
    res['msg'] = 'success'
    return res

@app.post("/query_runtime/{tester}")
async def query_runtime(tester: str, item: Item):
    item_dict = item.dict()
    config_body = {}
    config_body['tester'] = str(tester)
    config_body['operation_time'] = item_dict['operation_time']
    config_body['query_type'] = item_dict['query_type']
    config_body['category_item'] = item_dict['category_item']
    config_body['operation_item'] = item_dict['operation_item']
    output_body = {}
    output_body = dwave_leap.query_ising_model_runtime(config_body['query_type'])
    input_body = {}
    input_body = output_body['timing']
    input_body['problem_id'] = output_body['problem_id']
    res = input_body

    if item_dict['if_mq'] == 'True':
        mq_dwave_runtime.mq_write(config_body, input_body)
        res['mq_write_state'] = True
    else:
        res['mq_write_state'] = False
    res['code'] = 200
    res['msg'] = 'success'
    return res


if __name__ == '__main__':
    uvicorn.run(app='agent_api:app', host = '0.0.0.0', port = 1981, reload= True, debug = True)






# -*- coding:utf-8 -*-
from dwave.cloud import Client
import json
from dwave.system import DWaveSampler, EmbeddingComposite, LeapHybridSampler
import random
# from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from datahub.models import FieldType, RecordSchema, TupleRecord, BlobRecord, CursorType, RecordType
from datahub.exceptions import ResourceExistException, DatahubException
from datahub import DataHub
import sys
import time
import traceback
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

class Dwave_leap():
    def __init__(self, TOKEN):
        self.client = Client.from_config(token = TOKEN)
        print('*' * 50)
        print('Successfully acess to Dwave Leap cloud platform!')
        print('*' * 50)
#        self.qpu_advantage_sampler_embedding = EmbeddingComposite(DWaveSampler(solver={'qpu': True}))
        self.qpu_advantage = DWaveSampler(solver={'topology__type': 'pegasus'})
        self.qpu_advantage_sampler_embedding = EmbeddingComposite(self.qpu_advantage)        
        self.qpu_2000q = DWaveSampler(solver={'topology__type': 'chimera'})
        self.qpu_2000q_sampler_embedding = EmbeddingComposite(self.qpu_2000q)
        self.hybrid_bqm_sampler = LeapHybridSampler()        
        print('--------------successfully load!!------------------')
    def query_solver(self):
        return self.client.get_solvers()
    
    def query_ising_model_runtime(self,query_type = 'qpu_2000q'):
        if query_type == 'qpu_2000q':
            sampleset = self.qpu_2000q_sampler_embedding.sample_ising({'a': 1}, {('a', 'b'): 1})
        elif query_type == 'qpu_advantage':
            sampleset = self.qpu_advantage_sampler_embedding.sample_ising({'a': 1}, {('a', 'b'): 1})            
        else:
            sampleset = {}                   
        print('runtime of runing 2 items of Isring model is:\n')
        print(sampleset.info['timing'])
        return sampleset.info
    
    def query_solver_property(self, query_type = 'qpu_2000q'):
        if query_type == 'qpu_2000q':
            res_2000q = {}
            keys = self.qpu_2000q.properties.keys()
            for key in keys:
                res_2000q[key] = self.qpu_2000q.properties[key]
            return res_2000q
        elif query_type == 'qpu_advantage':
            res_qpu_advantage = {}
            keys = self.qpu_advantage.properties.keys()
            for key in keys:
                res_qpu_advantage[key] = self.qpu_advantage.properties[key]
            return res_qpu_advantage
        elif query_type == 'hybrid_bqm_sampler':
            res_hybrid_bqm_sampler = {}
            keys = self.hybrid_bqm_sampler.properties.keys()
            for key in keys:
                res_hybrid_bqm_sampler[key] = self.hybrid_bqm_sampler.properties[key]
            return res_hybrid_bqm_sampler
        else:
            return {}

class MQ_Dwave_Leap_Runtime:
    def __init__(self):
        access_id = ''
        access_key = ''
        endpoint = 'https://dh-cn-beijing.aliyuncs.com'
        self.dh = DataHub(access_id, access_key, endpoint)
        project_name = 'qc_benchmark_dwave_leap'
        comment = 'caict qc_benchmark testing platform'
        try:
            self.dh.create_project(project_name, comment)
            print("create project success!")
            print("=======================================\n\n")
            
        except ResourceExistException:
            print("project already exist!")
            print("=======================================\n\n")
        except Exception as e:
            print(traceback.format_exc())
            sys.exit(-1)

        # --------------------- topic ---------------------
        tuple_topic = "query_ising_model_runtime"
        shard_count = 3
        life_cycle = 7
        record_schema = RecordSchema.from_lists(
            ['tester', 'operation_time', 'query_type','category_item', 'operation_item', 'qpu_sampling_time',
             'qpu_anneal_time_per_sample', 'qpu_readout_time_per_sample',
             'qpu_access_time', 'qpu_access_overhead_time', 'qpu_programming_time',
             'qpu_delay_time_per_sample', 'total_post_processing_time', 'post_processing_overhead_time', 'problem_id'],
            [FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING,
             FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING,
             FieldType.STRING])

        try:
            self.dh.create_tuple_topic(project_name, tuple_topic,
                                       shard_count, life_cycle, record_schema, comment)
            print("create detection topic success!")
            print("=======================================\n\n")
        except ResourceExistException:
            print("topic already exist!")
            print("=======================================\n\n")
        except Exception as e:
            print(traceback.format_exc())
            sys.exit(-1)

        # ===================== get topic =====================
        topic_result = self.dh.get_topic(project_name, tuple_topic)
        print(topic_result)
        print(topic_result.record_schema)

        # ===================== list shard =====================
        shards_result = self.dh.list_shard(project_name, tuple_topic)
        print(shards_result)
        self.project_name = project_name
        self.tuple_topic = tuple_topic
        self.record_schema = record_schema

        self.output_body = {}
        self.output_body['tester'] = 'lvbo'
        self.output_body['operation_time'] = '1970-01-01 00:00:00'
        self.output_body['query_type'] = 'qpu_advatange'
        self.output_body['category_item'] = 'operation_maintenance'


    def mq_write(self, config_body, input_body):

        try:
            self.dh.wait_shards_ready(self.project_name, self.tuple_topic)
            # print("shards all ready!!!")
            # print("=======================================\n\n")

            topic_result = self.dh.get_topic(
                self.project_name, self.tuple_topic)
            #        print(topic_result)
            if topic_result.record_type != RecordType.TUPLE:
                print("topic type illegal!")
                sys.exit(-1)
            #        print("=======================================\n\n")

            #record_schema = topic_result.record_schema
            #
            records0 = []
            record1 = TupleRecord(schema = self.record_schema)
            keys = config_body.keys()
            for key in keys:
                record1.set_value(str(key), str(config_body[key]))
            key_words = input_body.keys()
            for key in key_words:
                record1.set_value(str(key), str(input_body[key]))           
            record1.hash_key = '7FFFFFFFFFFFFFFD7FFFFFFFFFFFFFFD'
            records0.append(record1)

            self.dh.put_records(self.project_name, self.tuple_topic, records0)
            print('*'*30)
            print('successfully write to datahub(MQ)!!')
        except DatahubException as e:
            print(e)
            sys.exit(-1)

class MQ_Dwave_Leap_Sovler_property:
    def __init__(self):
        access_id = ''
        access_key = ''
        endpoint = 'https://dh-cn-beijing.aliyuncs.com'
        self.dh = DataHub(access_id, access_key, endpoint)
        project_name = 'qc_benchmark_dwave_leap'
        comment = 'caict qc_benchmark testing platform'
        try:
            self.dh.create_project(project_name, comment)
            print("create project success!")
            print("=======================================\n\n")
        except ResourceExistException:
            print("project already exist!")
            print("=======================================\n\n")
        except Exception as e:
            print(traceback.format_exc())
            sys.exit(-1)

        # --------------------- topic ---------------------
        tuple_topic = "query_solver_property"
        shard_count = 3
        life_cycle = 7
        record_schema = RecordSchema.from_lists(
            ['tester', 'operation_time', 'query_type','category_item', 'operation_item', 'num_qubits', 'qubits', 'couplers',
             'h_range', 'j_range', 'supported_problem_types', 'parameters', 'vfyc', 'anneal_offset_ranges', 'anneal_offset_step',
             'anneal_offset_step_phi0', 'annealing_time_range', 'chip_id', 'default_annealing_time', 'default_programming_thermalization',
             'default_readout_thermalization', 'extended_j_range', 'h_gain_schedule_range', 'max_anneal_schedule_points',
             'max_h_gain_schedule_points', 'num_reads_range', 'per_qubit_coupling_range', 'problem_run_duration_range',
             'programming_thermalization_range', 'readout_thermalization_range', 'tags', 'topology', 'category', 'quota_conversion_rate'],
            [FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING,
             FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING,
             FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING,
             FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING,
             FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING, FieldType.STRING])

        try:
            self.dh.create_tuple_topic(project_name, tuple_topic,
                                       shard_count, life_cycle, record_schema, comment)
            print("create detection topic success!")
            print("=======================================\n\n")
        except ResourceExistException:
            print("topic already exist!")
            print("=======================================\n\n")
        except Exception as e:
            print(traceback.format_exc())
            sys.exit(-1)

        # ===================== get topic =====================
        topic_result = self.dh.get_topic(project_name, tuple_topic)
        print(topic_result)
        print(topic_result.record_schema)

        # ===================== list shard =====================
        shards_result = self.dh.list_shard(project_name, tuple_topic)
        print(shards_result)
        self.project_name = project_name
        self.tuple_topic = tuple_topic
        self.record_schema = record_schema

        self.output_body = {}
        self.output_body['tester'] = 'lvbo'
        self.output_body['operation_time'] = '1970-01-01 00:00:00'
        self.output_body['query_type'] = 'qpu_2000q'
        self.output_body['category_item'] = 'operation_maintenance'


    def mq_write(self, config_body, input_body):

        try:
            self.dh.wait_shards_ready(self.project_name, self.tuple_topic)
            # print("shards all ready!!!")
            # print("=======================================\n\n")

            topic_result = self.dh.get_topic(
                self.project_name, self.tuple_topic)
            #        print(topic_result)
            if topic_result.record_type != RecordType.TUPLE:
                print("topic type illegal!")
                sys.exit(-1)
            #        print("=======================================\n\n")

            #record_schema = topic_result.record_schema
            #
            records0 = []
            record1 = TupleRecord(schema = self.record_schema)
            keys = config_body.keys()
            for key in keys:
                record1.set_value(str(key), str(config_body[key]))
            key_words = input_body.keys()
            for key in key_words:
                record1.set_value(str(key), str(input_body[key]))
            record1.hash_key = '6FFFFFFFFFFFFFFD7FFFFFFFFFFFFFFD'
            records0.append(record1)

            self.dh.put_records(self.project_name, self.tuple_topic, records0)
            print('*'*30)
            print('successfully write to datahub(MQ)!!')
        except DatahubException as e:
            print(e)
            sys.exit(-1)

if  __name__ == '__main__':
    config_body = {}
    config_body['tester'] = 'lvbo-caict'
    config_body['operation_time'] = 'cc'
    config_body['query_type'] = 'qpu_advantage'
    config_body['category_item'] = 'operation_maintenance'
    TOKEN = ''
    dwave_leap = Dwave_leap(TOKEN)
    mq_dwave_solver = MQ_Dwave_Leap_Sovler_property()
    mq_dwave_runtime = MQ_Dwave_Leap_Runtime()
    output_body = {}
    output_body = dwave_leap.query_ising_model_runtime(config_body['query_type'])
    input_body = {}
    input_body = output_body['timing']
    input_body['problem_id'] = output_body['problem_id']
    time.sleep(2)
    mq_dwave_runtime.mq_write(config_body, input_body)

